import pytest

from swerex.deployment.config import DockerDeploymentConfig
from swerex.deployment.docker import DockerDeployment
from swerex.utils.free_port import find_free_port


async def test_docker_deployment():
    port = find_free_port()
    print(f"Using port {port} for the docker deployment")
    d = DockerDeployment(image="swe-rex-test:latest", port=port)
    with pytest.raises(RuntimeError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()


@pytest.mark.slow
async def test_docker_deployment_with_python_standalone():
    port = find_free_port()
    print(f"Using port {port} for the docker deployment")
    d = DockerDeployment(image="ubuntu:latest", port=port, python_standalone_dir="/root")
    with pytest.raises(RuntimeError):
        await d.is_alive()
    await d.start()
    assert await d.is_alive()
    await d.stop()


@pytest.mark.slow
def test_docker_deployment_config_platform():
    config = DockerDeploymentConfig(docker_args=["--platform", "linux/amd64", "--other-arg"])
    assert config.platform == "linux/amd64"

    config = DockerDeploymentConfig(docker_args=["--platform=linux/amd64", "--other-arg"])
    assert config.platform == "linux/amd64"

    config = DockerDeploymentConfig(docker_args=["--other-arg"])
    assert config.platform is None

    with pytest.raises(ValueError):
        config = DockerDeploymentConfig(platform="linux/amd64", docker_args=["--platform", "linux/amd64"])
    with pytest.raises(ValueError):
        config = DockerDeploymentConfig(platform="linux/amd64", docker_args=["--platform=linux/amd64"])


def test_docker_deployment_config_container_runtime():
    # Test default container runtime is docker
    config = DockerDeploymentConfig(image="test")
    assert config.container_runtime == "docker"

    # Test setting container runtime to podman
    config = DockerDeploymentConfig(image="test", container_runtime="podman")
    assert config.container_runtime == "podman"


async def test_podman_deployment():
    """Test deployment with Podman container runtime"""
    port = find_free_port()
    print(f"Using port {port} for the podman deployment")
    d = DockerDeployment(image="swe-rex-test:latest", port=port, container_runtime="podman")
    assert d._config.container_runtime == "podman"
    # Note: This test will only pass if podman is installed and the test image exists
    # In CI/CD environments without podman, this test may be skipped


def test_podman_deployment_config():
    """Test that DockerDeployment works with podman configuration"""
    config = DockerDeploymentConfig(image="test:latest", container_runtime="podman", port=8080, pull="never")
    deployment = DockerDeployment.from_config(config)
    assert deployment._config.container_runtime == "podman"
    assert deployment._config.image == "test:latest"
    assert deployment._config.port == 8080


def test_docker_deployment_config_container_runtime():
    # Test default container runtime is docker
    config = DockerDeploymentConfig(image="test")
    assert config.container_runtime == "docker"

    # Test setting container runtime to podman
    config = DockerDeploymentConfig(image="test", container_runtime="podman")
    assert config.container_runtime == "podman"


async def test_podman_deployment():
    """Test deployment with Podman container runtime"""
    port = find_free_port()
    print(f"Using port {port} for the podman deployment")
    d = DockerDeployment(image="swe-rex-test:latest", port=port, container_runtime="podman")
    assert d._config.container_runtime == "podman"
    # Note: This test will only pass if podman is installed and the test image exists
    # In CI/CD environments without podman, this test may be skipped


def test_podman_deployment_config():
    """Test that DockerDeployment works with podman configuration"""
    config = DockerDeploymentConfig(image="test:latest", container_runtime="podman", port=8080, pull="never")
    deployment = DockerDeployment.from_config(config)
    assert deployment._config.container_runtime == "podman"
    assert deployment._config.image == "test:latest"
    assert deployment._config.port == 8080



import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from botocore.waiter import WaiterError

"""
AWS Resource Cleanup Utility

This script automates the cleanup of AWS resources that were created by the SWE-REX AWS deployment process.
All resources are tagged with the 'origin=swe-rex-deployment-auto' tag to identify them as part of the SWE-REX deployment. These resources are removed by this script.

Resources handled:
- ECS Clusters (including running tasks)
- ECS Task Definitions
- Security Groups
- IAM Roles

Usage:
    python -m swerex.utils.aws_teardown

The script will preview all resources to be deleted and request confirmation before proceeding.
Each step is handled independently, allowing for partial cleanup if errors occur.
"""


def get_confirmation(message: str) -> bool:
    """
    Request user confirmation before proceeding with destructive actions.

    Args:
        message: The confirmation message to display to the user

    Returns:
        bool: True if user confirms, False otherwise
    """
    response = input(f"\n{message}\nDo you want to continue? (y/n): ").lower().strip()
    return response == "y"


def has_target_tag(tags: list[dict]) -> bool:
    """
    Check if a resource has the SWE-REX deployment tag.

    Args:
        tags: List of tag dictionaries from AWS resource

    Returns:
        bool: True if the target tag is present
    """
    return any(
        tag.get("key", tag.get("Key", "")) == "origin"
        and tag.get("value", tag.get("Value", "")) == "swe-rex-deployment-auto"
        for tag in tags
    )


def delete_iam_roles() -> None:
    """
    Delete all tagged IAM roles and their attached policies.
    Handles pagination for large numbers of roles.
    """
    iam = boto3.client("iam")
    paginator = iam.get_paginator("list_roles")

    # Preview roles to be deleted
    roles_to_delete = []
    for page in paginator.paginate():
        for role in page["Roles"]:
            tags = iam.list_role_tags(RoleName=role["RoleName"])["Tags"]
            if has_target_tag(tags):
                roles_to_delete.append(role["RoleName"])

    if not roles_to_delete:
        print("No IAM roles found to delete.")
        return

    confirmation_msg = "The following IAM roles will be deleted:\n" + "\n".join(f"- {role}" for role in roles_to_delete)
    if not get_confirmation(confirmation_msg):
        print("Skipping IAM role deletion.")
        return

    for role_name in roles_to_delete:
        try:
            print(f"Cleaning up IAM role: {role_name}")
            iam.delete_role(RoleName=role_name)
        except ClientError as e:
            print(f"Error processing role {role_name}: {str(e)}")


def delete_task_definitions() -> None:
    """
    Deregister all tagged ECS task definitions.
    Task definitions are not actually deleted but marked as INACTIVE.
    """
    ecs = boto3.client("ecs")
    paginator = ecs.get_paginator("list_task_definitions")

    # Preview task definitions to be deleted
    tasks_to_delete = []
    for page in paginator.paginate():
        for task_def_arn in page["taskDefinitionArns"]:
            tags = ecs.list_tags_for_resource(resourceArn=task_def_arn)["tags"]
            if has_target_tag(tags):
                tasks_to_delete.append(task_def_arn)

    if not tasks_to_delete:
        print("No task definitions found to delete.")
        return

    confirmation_msg = "The following task definitions will be deregistered:\n" + "\n".join(
        f"- {task}" for task in tasks_to_delete
    )
    if not get_confirmation(confirmation_msg):
        print("Skipping task definition deletion.")
        return

    for task_def_arn in tasks_to_delete:
        try:
            print(f"Deregistering task definition: {task_def_arn}")
            ecs.deregister_task_definition(taskDefinition=task_def_arn)
        except ClientError as e:
            print(f"Error processing task definition {task_def_arn}: {str(e)}")


def delete_ecs_clusters() -> None:
    """
    Delete all tagged ECS clusters and their running tasks.
    Waits for tasks to stop before attempting cluster deletion.
    """
    ecs = boto3.client("ecs")

    # Preview clusters to be deleted
    clusters = ecs.list_clusters()["clusterArns"]
    clusters_to_delete = []
    for cluster_arn in clusters:
        tags = ecs.list_tags_for_resource(resourceArn=cluster_arn)["tags"]
        if has_target_tag(tags):
            clusters_to_delete.append(cluster_arn)

    if not clusters_to_delete:
        print("No ECS clusters found to delete.")
        return

    confirmation_msg = "The following ECS clusters will be deleted:\n" + "\n".join(
        f"- {cluster}" for cluster in clusters_to_delete
    )
    if not get_confirmation(confirmation_msg):
        print("Skipping ECS cluster deletion.")
        return

    for cluster_arn in clusters_to_delete:
        try:
            print(f"Deleting ECS cluster: {cluster_arn}")
            # Stop all tasks in the cluster using pagination
            paginator = ecs.get_paginator("list_tasks")
            for page in paginator.paginate(cluster=cluster_arn):
                tasks = page["taskArns"]
                for task in tasks:
                    print(f"Stopping task: {task}")
                    ecs.stop_task(cluster=cluster_arn, task=task)

            # Wait for tasks to stop before deleting cluster
            waiter = ecs.get_waiter("tasks_stopped")
            try:
                waiter.wait(cluster=cluster_arn)
            except WaiterError:
                print(f"Warning: Timeout waiting for tasks to stop in cluster {cluster_arn}")

            # Delete the cluster
            ecs.delete_cluster(cluster=cluster_arn)
        except ClientError as e:
            print(f"Error processing cluster {cluster_arn}: {str(e)}")


def delete_security_groups() -> None:
    """
    Delete all tagged security groups.
    Removes all ingress and egress rules before deletion to handle dependencies.
    """
    ec2 = boto3.client("ec2")

    # Preview security groups to be deleted
    security_groups = ec2.describe_security_groups()["SecurityGroups"]
    sgs_to_delete = []
    for sg in security_groups:
        tags = sg.get("Tags", [])
        if has_target_tag(tags):
            sgs_to_delete.append(f"{sg['GroupId']} ({sg.get('GroupName', 'No Name')})")

    if not sgs_to_delete:
        print("No security groups found to delete.")
        return

    confirmation_msg = "The following security groups will be deleted:\n" + "\n".join(f"- {sg}" for sg in sgs_to_delete)
    if not get_confirmation(confirmation_msg):
        print("Skipping security group deletion.")
        return

    for sg in security_groups:
        try:
            if has_target_tag(sg.get("Tags", [])):
                print(f"Deleting security group: {sg['GroupId']}")
                # Remove all inbound rules first
                if sg.get("IpPermissions"):
                    ec2.revoke_security_group_ingress(GroupId=sg["GroupId"], IpPermissions=sg["IpPermissions"])
                # Remove all outbound rules
                if sg.get("IpPermissionsEgress"):
                    ec2.revoke_security_group_egress(GroupId=sg["GroupId"], IpPermissions=sg["IpPermissionsEgress"])
                # Delete the security group
                ec2.delete_security_group(GroupId=sg["GroupId"])
        except ClientError as e:
            print(f"Error processing security group {sg['GroupId']}: {str(e)}")


def main():
    """
    Orchestrate the cleanup of all AWS resources tagged with origin=swe-rex-deployment-auto.

    Process:
    1. Configure AWS clients with appropriate timeouts
    2. Preview all resources that will be affected
    3. Request user confirmation
    4. Delete resources in dependency order:
       - ECS clusters (and tasks)
       - Task definitions
       - Security groups
       - IAM roles
    """
    print("Starting cleanup of AWS resources tagged with origin=swe-rex-deployment-auto; this may take a while...")

    try:
        # Fix the Config instantiation
        config = Config(connect_timeout=5, read_timeout=10, retries={"max_attempts": 2})
        ecs = boto3.client("ecs", config=config)
        iam = boto3.client("iam", config=config)
        ec2 = boto3.client("ec2", config=config)

        # Get clusters and their tasks using pagination with timeout
        clusters = []
        try:
            clusters = ecs.list_clusters()["clusterArns"]
        except ClientError as e:
            print(f"Warning: Failed to list ECS clusters: {e}")
            clusters = []

        cluster_info = []
        for cluster_arn in clusters:
            try:
                tags = ecs.list_tags_for_resource(resourceArn=cluster_arn)["tags"]
                if has_target_tag(tags):
                    task_count = 0
                    paginator = ecs.get_paginator("list_tasks")
                    for page in paginator.paginate(cluster=cluster_arn, PaginationConfig={"MaxItems": 100}):
                        task_count += len(page["taskArns"])
                    cluster_name = cluster_arn.split("/")[-1]
                    cluster_info.append(f"  - Cluster: {cluster_name} ({task_count} running tasks)")
            except ClientError as e:
                print(f"Warning: Failed to process cluster {cluster_arn}: {e}")
                continue

        # Get task definitions
        task_defs = []
        paginator = ecs.get_paginator("list_task_definitions")
        for page in paginator.paginate():
            for task_def_arn in page["taskDefinitionArns"]:
                tags = ecs.list_tags_for_resource(resourceArn=task_def_arn)["tags"]
                if has_target_tag(tags):
                    task_name = task_def_arn.split("/")[-1]
                    task_defs.append(f"  - {task_name}")

        # Get IAM roles
        roles = []
        paginator = iam.get_paginator("list_roles")
        for page in paginator.paginate():
            for role in page["Roles"]:
                tags = iam.list_role_tags(RoleName=role["RoleName"])["Tags"]
                if has_target_tag(tags):
                    roles.append(f"  - {role['RoleName']}")

        # Get security groups
        security_groups = []
        for sg in ec2.describe_security_groups()["SecurityGroups"]:
            tags = sg.get("Tags", [])
            if has_target_tag(tags):
                security_groups.append(f"  - {sg['GroupId']} ({sg.get('GroupName', 'No Name')})")

        # Build detailed confirmation message
        confirmation_msg = "The following AWS resources will be deleted:\n\n"

        if cluster_info:
            confirmation_msg += f"ECS Clusters ({len(cluster_info)}):\n" + "\n".join(cluster_info) + "\n\n"

        if task_defs:
            confirmation_msg += f"Task Definitions ({len(task_defs)}):\n" + "\n".join(task_defs) + "\n\n"

        if security_groups:
            confirmation_msg += f"Security Groups ({len(security_groups)}):\n" + "\n".join(security_groups) + "\n\n"

        if roles:
            confirmation_msg += f"IAM Roles ({len(roles)}):\n" + "\n".join(roles) + "\n\n"

        confirmation_msg += "This action cannot be undone."

        if not get_confirmation(confirmation_msg):
            print("Cleanup cancelled.")
            return

        # Order matters due to dependencies
        print("\nStopping and deleting ECS tasks and clusters...")
        delete_ecs_clusters()

        print("\nDeregistering task definitions...")
        delete_task_definitions()

        print("\nDeleting security groups...")
        delete_security_groups()

        print("\nCleaning up IAM roles...")
        delete_iam_roles()

        print("\nCleanup complete!")
    except Exception as e:
        print(f"Error during cleanup: {str(e)}")


if __name__ == "__main__":
    main()



import logging
import time
import uuid
from typing import Any

from daytona_sdk import (
    CreateSandboxFromImageParams,
    Daytona,
    DaytonaConfig,
    SessionExecuteRequest,
)
from typing_extensions import Self

from swerex import PACKAGE_NAME, REMOTE_EXECUTABLE_NAME
from swerex.deployment.abstract import AbstractDeployment
from swerex.deployment.config import DaytonaDeploymentConfig
from swerex.deployment.hooks.abstract import CombinedDeploymentHook, DeploymentHook
from swerex.exceptions import DeploymentNotStartedError
from swerex.runtime.abstract import IsAliveResponse
from swerex.runtime.remote import RemoteRuntime
from swerex.utils.log import get_logger
from swerex.utils.wait import _wait_until_alive


class DaytonaDeployment(AbstractDeployment):
    def __init__(
        self,
        *,
        logger: logging.Logger | None = None,
        **kwargs: Any,
    ):
        self._config = DaytonaDeploymentConfig(**kwargs)
        self._runtime: RemoteRuntime | None = None
        self._sandbox = None
        self._sandbox_id = None
        self.logger = logger or get_logger("rex-deploy")
        self._hooks = CombinedDeploymentHook()
        self._daytona = None
        self._auth_token = None

    def add_hook(self, hook: DeploymentHook):
        self._hooks.add_hook(hook)

    @classmethod
    def from_config(cls, config: DaytonaDeploymentConfig) -> Self:
        return cls(**config.model_dump())

    def _init_daytona(self):
        """Initialize the Daytona client with configuration."""
        daytona_config = DaytonaConfig(api_key=self._config.api_key, target=self._config.target)
        self._daytona = Daytona(daytona_config)

    def _get_token(self) -> str:
        """Generate a unique authentication token."""
        return str(uuid.uuid4())

    def _get_command(self, *, token: str) -> str:
        """Generate the command to run the SWE Rex server."""
        main_command = f"{REMOTE_EXECUTABLE_NAME} --port {self._config.port} --auth-token {token}"
        fallback_commands = [
            "apt-get update -y",
            "apt-get install pipx -y",
            "pipx ensurepath",
            f"pipx run {PACKAGE_NAME} --port {self._config.port} --auth-token {token}",
        ]
        fallback_script = " && ".join(fallback_commands)
        # Wrap the entire command in bash -c to ensure timeout applies to everything
        inner_command = f"{main_command} || ( {fallback_script} )"
        return f"timeout {self._config.container_timeout}s bash -c '{inner_command}'"

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None or self._sandbox is None:
            raise DeploymentNotStartedError()

        # Check if the workspace is still running
        try:
            # Get session status to verify the workspace is still active
            sessions = self._sandbox.process.list_sessions()
            if not sessions:
                msg = "Daytona workspace has no active sessions"
                raise RuntimeError(msg)
        except Exception as e:
            msg = f"Error checking Daytona workspace status: {str(e)}"
            raise RuntimeError(msg)

        return await self._runtime.is_alive(timeout=timeout)

    async def _wait_until_alive(self, timeout: float):
        """Wait until the runtime is alive."""
        return await _wait_until_alive(self.is_alive, timeout=timeout, function_timeout=self._config.container_timeout)

    async def start(self):
        """Starts the runtime in a Daytona sandbox."""
        self._init_daytona()
        self.logger.info("Creating Daytona sandbox...")

        # Create workspace with specified parameters
        params = CreateSandboxFromImageParams(
            image=self._config.image,
        )
        assert self._daytona is not None

        self._sandbox = self._daytona.create(params)
        self._sandbox_id = self._sandbox.id
        self.logger.info(f"Created Daytona sandbox with ID: {self._sandbox_id}")

        # Generate authentication token
        self._auth_token = self._get_token()

        # Run the SWE Rex server in the sandbox
        command = self._get_command(token=self._auth_token)
        self.logger.info("Starting SWE Rex server in Daytona sandbox...")

        # Create a session for the long-running process
        session_id = f"swerex-server-{uuid.uuid4().hex[:8]}"
        self._sandbox.process.create_session(session_id)

        req = SessionExecuteRequest(command=command, runAsync=True)
        # Execute the command in the session
        response = self._sandbox.process.execute_session_command(session_id, req)
        if response.exit_code != 0:
            self.logger.error(f"Failed to start SWE Rex server: {response.output}")
            await self.stop()
            msg = f"Failed to start SWE Rex server: {response.output}"
            raise RuntimeError(msg)

        sweRexHost = self._sandbox.get_preview_link(self._config.port)

        # Create the remote runtime
        self._runtime = RemoteRuntime(host=sweRexHost, port=None, auth_token=self._auth_token, logger=self.logger)

        # Wait for the runtime to be alive
        t0 = time.time()
        await self._wait_until_alive(timeout=self._config.runtime_timeout)
        self.logger.info(f"Runtime started in {time.time() - t0:.2f}s")

    async def stop(self):
        """Stops the runtime and removes the Daytona sandbox."""
        if self._runtime is not None:
            await self._runtime.close()
            self._runtime = None

        if self._sandbox is not None and self._daytona is not None:
            try:
                self.logger.info(f"Removing Daytona sandbox with ID: {self._sandbox_id}")
                self._daytona.delete(self._sandbox)
                self.logger.info("Daytona sandbox removed successfully")
            except Exception as e:
                self.logger.error(f"Failed to remove Daytona sandbox: {str(e)}")

        self._sandbox = None
        self._sandbox_id = None
        self._auth_token = None

    @property
    def runtime(self) -> RemoteRuntime:
        """Returns the runtime if running.

        Raises:
            DeploymentNotStartedError: If the deployment was not started.
        """
        if self._runtime is None:
            msg = "Runtime not started"
            raise RuntimeError(msg)
        return self._runtime



import asyncio
import logging
import random
import shutil
import sys
import tempfile
import traceback
import uuid
from pathlib import Path
from typing import Any

import aiohttp
from pydantic import BaseModel
from typing_extensions import Self

from swerex.exceptions import SwerexException
from swerex.runtime.abstract import (
    AbstractRuntime,
    Action,
    CloseResponse,
    CloseSessionRequest,
    CloseSessionResponse,
    Command,
    CommandResponse,
    CreateSessionRequest,
    CreateSessionResponse,
    IsAliveResponse,
    Observation,
    ReadFileRequest,
    ReadFileResponse,
    UploadRequest,
    UploadResponse,
    WriteFileRequest,
    WriteFileResponse,
    _ExceptionTransfer,
)
from swerex.runtime.config import RemoteRuntimeConfig
from swerex.utils.log import get_logger
from swerex.utils.wait import _wait_until_alive

__all__ = ["RemoteRuntime", "RemoteRuntimeConfig"]


class RemoteRuntime(AbstractRuntime):
    def __init__(
        self,
        *,
        logger: logging.Logger | None = None,
        **kwargs: Any,
    ):
        """A runtime that connects to a remote server.

        Args:
            **kwargs: Keyword arguments to pass to the `RemoteRuntimeConfig` constructor.
        """
        self._config = RemoteRuntimeConfig(**kwargs)
        self.logger = logger or get_logger("rex-runtime")
        if not self._config.host.startswith("http"):
            self.logger.warning("Host %s does not start with http, adding http://", self._config.host)
            self._config.host = f"http://{self._config.host}"

    @classmethod
    def from_config(cls, config: RemoteRuntimeConfig) -> Self:
        return cls(**config.model_dump())

    def _get_timeout(self, timeout: float | None = None) -> float:
        if timeout is None:
            return self._config.timeout
        return timeout

    @property
    def _headers(self) -> dict[str, str]:
        """Request headers to use for authentication."""
        if self._config.auth_token:
            return {"X-API-Key": self._config.auth_token}
        return {}

    @property
    def _api_url(self) -> str:
        if self._config.port is None:
            return self._config.host
        return f"{self._config.host}:{self._config.port}"

    def _handle_transfer_exception(self, exc_transfer: _ExceptionTransfer) -> None:
        """Reraise exceptions that were thrown on the remote."""
        if exc_transfer.traceback:
            self.logger.critical("Traceback: \n%s", exc_transfer.traceback)
        module, _, exc_name = exc_transfer.class_path.rpartition(".")
        print(module, exc_name)
        if module == "builtins":
            module_obj = __builtins__
        else:
            if module not in sys.modules:
                self.logger.debug("Module %s not in sys.modules, trying to import it", module)
                try:
                    __import__(module)
                except ImportError:
                    self.logger.debug("Failed to import module %s", module)
                    exc = SwerexException(exc_transfer.message)
                    raise exc from None
            module_obj = sys.modules[module]
        try:
            if isinstance(module_obj, dict):
                # __builtins__, sometimes
                exception = module_obj[exc_name](exc_transfer.message)
            else:
                exception = getattr(module_obj, exc_name)(exc_transfer.message)
        except (AttributeError, TypeError):
            self.logger.error(
                f"Could not initialize transferred exception: {exc_transfer.class_path!r}. "
                f"Transfer object: {exc_transfer}"
            )
            exception = SwerexException(exc_transfer.message)
        exception.extra_info = exc_transfer.extra_info
        raise exception from None

    async def _handle_response_errors(self, response: aiohttp.ClientResponse) -> None:
        """Raise exceptions found in the request response."""
        if response.status == 511:
            data = await response.json()
            exc_transfer = _ExceptionTransfer(**data["swerexception"])
            self._handle_transfer_exception(exc_transfer)
        if response.status >= 400:
            data = await response.json()
            self.logger.critical("Received error response: %s", data)
            response.raise_for_status()

    async def is_alive(self, *, timeout: float | None = None) -> IsAliveResponse:
        """Checks if the runtime is alive.

        Internal server errors are thrown, everything else just has us return False
        together with the message.
        """
        try:
            async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(force_close=True)) as session:
                timeout_value = self._get_timeout(timeout)
                async with session.get(
                    f"{self._api_url}/is_alive",
                    headers=self._headers,
                    timeout=aiohttp.ClientTimeout(total=timeout_value),
                ) as response:
                    if response.status == 200:
                        data = await response.json()
                        return IsAliveResponse(**data)
                    elif response.status == 511:
                        data = await response.json()
                        exc_transfer = _ExceptionTransfer(**data["swerexception"])
                        self._handle_transfer_exception(exc_transfer)

                    data = await response.json()
                    msg = f"Status code {response.status} from {self._api_url}/is_alive. Message: {data.get('detail')}"
                    return IsAliveResponse(is_alive=False, message=msg)
        except aiohttp.ClientError:
            msg = f"Failed to connect to {self._config.host}\n"
            msg += traceback.format_exc()
            return IsAliveResponse(is_alive=False, message=msg)
        except Exception:
            msg = f"Failed to connect to {self._config.host}\n"
            msg += traceback.format_exc()
            return IsAliveResponse(is_alive=False, message=msg)

    async def wait_until_alive(self, *, timeout: float = 60.0):
        return await _wait_until_alive(self.is_alive, timeout=timeout)

    async def _request(self, endpoint: str, payload: BaseModel | None, output_class: Any, num_retries: int = 0):
        """Small helper to make requests to the server and handle errors and output."""
        request_url = f"{self._api_url}/{endpoint}"
        request_id = str(uuid.uuid4())
        headers = self._headers.copy()
        headers["X-Request-ID"] = request_id  # idempotency key for the request

        retry_count = 0
        last_exception: Exception | None = None
        retry_delay = 0.1
        backoff_max = 5

        while retry_count <= num_retries:
            try:
                async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(force_close=True)) as session:
                    async with session.post(
                        request_url,
                        json=payload.model_dump() if payload else None,
                        headers=headers,
                    ) as resp:
                        await self._handle_response_errors(resp)
                        return output_class(**await resp.json())
            except Exception as e:
                last_exception = e
                retry_count += 1
                if retry_count <= num_retries:
                    await asyncio.sleep(retry_delay)
                    retry_delay *= 2
                    retry_delay += random.uniform(0, 0.5)
                    retry_delay = min(retry_delay, backoff_max)
                    continue
                self.logger.error("Error making request %s after %d retries: %s", request_id, num_retries, e)
        raise last_exception  # type: ignore

    async def create_session(self, request: CreateSessionRequest) -> CreateSessionResponse:
        """Creates a new session."""
        return await self._request("create_session", request, CreateSessionResponse)

    async def run_in_session(self, action: Action) -> Observation:
        """Runs a command in a session."""
        return await self._request("run_in_session", action, Observation)

    async def close_session(self, request: CloseSessionRequest) -> CloseSessionResponse:
        """Closes a shell session."""
        return await self._request("close_session", request, CloseSessionResponse)

    async def execute(self, command: Command) -> CommandResponse:
        """Executes a command (independent of any shell session)."""
        return await self._request("execute", command, CommandResponse)

    async def read_file(self, request: ReadFileRequest) -> ReadFileResponse:
        """Reads a file"""
        return await self._request("read_file", request, ReadFileResponse)

    async def write_file(self, request: WriteFileRequest) -> WriteFileResponse:
        """Writes a file"""
        return await self._request("write_file", request, WriteFileResponse)

    async def upload(self, request: UploadRequest) -> UploadResponse:
        """Uploads a file"""
        source = Path(request.source_path).resolve()
        self.logger.debug("Uploading file from %s to %s", request.source_path, request.target_path)

        async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(force_close=True)) as session:
            if source.is_dir():
                # Ignore cleanup errors: See https://github.com/SWE-agent/SWE-agent/issues/1005
                with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
                    zip_path = Path(temp_dir) / "zipped_transfer.zip"
                    shutil.make_archive(str(zip_path.with_suffix("")), "zip", source)
                    self.logger.debug("Created zip file at %s", zip_path)

                    with open(zip_path, "rb") as f:
                        data = aiohttp.FormData()
                        data.add_field("file", f, filename=zip_path.name, content_type="application/zip")
                        data.add_field("target_path", request.target_path)
                        data.add_field("unzip", "true")

                        async with session.post(
                            f"{self._api_url}/upload", data=data, headers=self._headers
                        ) as response:
                            await self._handle_response_errors(response)
                            return UploadResponse(**(await response.json()))
            elif source.is_file():
                self.logger.debug("Uploading file from %s to %s", source, request.target_path)

                with open(source, "rb") as f:
                    data = aiohttp.FormData()
                    data.add_field("file", f, filename=source.name)
                    data.add_field("target_path", request.target_path)
                    data.add_field("unzip", "false")

                    async with session.post(f"{self._api_url}/upload", data=data, headers=self._headers) as response:
                        await self._handle_response_errors(response)
                        return UploadResponse(**(await response.json()))
            else:
                msg = f"Source path {source} is not a file or directory"
                raise ValueError(msg)

    async def close(self) -> CloseResponse:
        """Closes the runtime."""
        return await self._request("close", None, CloseResponse)



class DeploymentHook:
    def on_custom_step(self, message: str): ...


class CombinedDeploymentHook(DeploymentHook):
    def __init__(self, hooks: list[DeploymentHook] | None = None):
        self._hooks = []
        for hook in hooks or []:
            self.add_hook(hook)

    def add_hook(self, hook: DeploymentHook):
        self._hooks.append(hook)

    def on_custom_step(self, message: str):
        for hook in self._hooks:
            hook.on_custom_step(message)



from collections.abc import Callable

from swerex.deployment.hooks.abstract import DeploymentHook


class SetStatusDeploymentHook(DeploymentHook):
    def __init__(self, id: str, callable: Callable[[str, str], None]):
        self._callable = callable
        self._id = id

    def _update(self, message: str):
        self._callable(self._id, message)  # type: ignore

    def on_custom_step(self, message: str):
        self._update(message)



import hashlib
import json
from urllib.parse import quote

import boto3


def get_name_hash(prefix: str, obj: dict, max_length: int = 128, hash_length: int = 12) -> str:
    prefix_length = min(max_length, len(prefix))
    if hash_length + prefix_length > max_length:
        msg = f"Prefix and hash length are too long: {prefix} has length {prefix_length}, hash length is {hash_length}, max length is {max_length}"
        raise ValueError(msg)
    return f"{prefix}-{hashlib.sha256(json.dumps(obj).encode()).hexdigest()[:hash_length]}"


def get_container_name(image_name: str) -> str:
    image_name_sanitized = "".join(c for c in image_name if c.isalnum() or c in "-_")
    if len(image_name_sanitized) <= 255:
        return image_name_sanitized
    return hashlib.sha256(image_name_sanitized.encode()).hexdigest()[:255]


def get_execution_role_arn(execution_role_prefix: str) -> str:
    iam_client = boto3.client("iam")

    trust_relationship = {
        "Version": "2012-10-17",
        "Statement": [
            {"Effect": "Allow", "Principal": {"Service": "ecs-tasks.amazonaws.com"}, "Action": "sts:AssumeRole"}
        ],
    }

    inline_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogGroup",
                    "logs:CreateLogStream",
                    "logs:PutLogEvents",
                    "logs:DescribeLogStreams",
                    "secretsmanager:GetSecretValue",
                ],
                "Resource": [
                    "arn:aws:logs:*:*:log-group:/ecs/swe-rex-deployment:*",
                    "arn:aws:logs:*:*:log-group:/ecs/swe-rex-deployment",
                ],
            }
        ],
    }
    role_name = get_name_hash(execution_role_prefix, inline_policy, max_length=64)
    # Check if the role already exists
    role_exists = False
    try:
        role = iam_client.get_role(RoleName=role_name)
        role_exists = True
    except iam_client.exceptions.NoSuchEntityException:
        pass

    if not role_exists:
        role = iam_client.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(trust_relationship),
            Description="Execution role for ECS tasks",
            MaxSessionDuration=3600,
            Tags=[{"Key": "origin", "Value": "swe-rex-deployment-auto"}],
        )
    # check if policies already attached
    attached_policies = iam_client.list_attached_role_policies(RoleName=role_name)
    attached_policy_arns = [policy["PolicyArn"] for policy in attached_policies["AttachedPolicies"]]

    if "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy" not in attached_policy_arns:
        iam_client.attach_role_policy(
            RoleName=role_name, PolicyArn="arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy"
        )
    if "arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly" not in attached_policy_arns:
        iam_client.attach_role_policy(
            RoleName=role_name, PolicyArn="arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly"
        )

    policy_names = iam_client.list_role_policies(RoleName=role_name)["PolicyNames"]
    if "LogsAndSecretsPolicy" not in policy_names:
        # Add inline policy
        iam_client.put_role_policy(
            RoleName=role_name, PolicyName="LogsAndSecretsPolicy", PolicyDocument=json.dumps(inline_policy)
        )
    waiter = iam_client.get_waiter("role_exists")
    waiter.wait(RoleName=role_name)
    return role["Role"]["Arn"]


def get_task_definition(
    image_name: str,
    port: int,
    execution_role_arn: str,
    task_definition_prefix: str,
    log_group: str | None = None,
) -> str:
    ecs_client = boto3.client("ecs")
    task_definition = {
        "executionRoleArn": execution_role_arn,
        "networkMode": "awsvpc",
        "memory": "2048",
        "cpu": "1 vCPU",
        "containerDefinitions": [
            {
                "name": get_container_name(image_name),
                "image": image_name,
                "portMappings": [{"containerPort": port, "hostPort": port, "protocol": "tcp"}],
                "essential": True,
                "entryPoint": ["/bin/sh", "-c"],
                "command": [
                    "echo 'hello world'"  # override command with run_task
                ],
            },
        ],
        "requiresCompatibilities": ["FARGATE", "EC2"],
    }
    if log_group:
        task_definition["containerDefinitions"][0]["logConfiguration"] = {
            "logDriver": "awslogs",
            "options": {
                "awslogs-group": log_group,
                "awslogs-region": "us-east-2",
                "awslogs-stream-prefix": "ecs",
                "awslogs-create-group": "true",
            },
        }
    family_name = get_name_hash(task_definition_prefix, task_definition, max_length=255)
    # if family exists just return the task definition
    try:
        response = ecs_client.describe_task_definition(taskDefinition=family_name)
        return response["taskDefinition"]
    except ecs_client.exceptions.ClientException:
        pass

    response = ecs_client.register_task_definition(
        family=family_name,
        **task_definition,
        tags=[{"key": "origin", "value": "swe-rex-deployment-auto"}],
    )
    return response["taskDefinition"]


def get_cluster_arn(cluster_name: str) -> str:
    ecs_client = boto3.client("ecs")
    response = ecs_client.create_cluster(
        clusterName=cluster_name,
        tags=[{"key": "origin", "value": "swe-rex-deployment-auto"}],
    )
    return response["cluster"]["clusterArn"]


def get_default_vpc_and_subnet() -> tuple[str, str]:
    ec2_client = boto3.client("ec2")
    vpcs = ec2_client.describe_vpcs(Filters=[{"Name": "isDefault", "Values": ["true"]}])
    if not vpcs["Vpcs"]:
        msg = "No default VPC found"
        raise Exception(msg)
    vpc_id = vpcs["Vpcs"][0]["VpcId"]
    subnets = ec2_client.describe_subnets(Filters=[{"Name": "vpc-id", "Values": [vpc_id]}])
    if not subnets["Subnets"]:
        msg = "No subnets found in the default VPC"
        raise Exception(msg)
    subnet_id = subnets["Subnets"][0]["SubnetId"]
    return vpc_id, subnet_id


def get_security_group(vpc_id: str, port: int, security_group_prefix: str) -> str:
    ec2_client = boto3.client("ec2")
    inbound_rule = {"IpProtocol": "tcp", "FromPort": port, "ToPort": port, "IpRanges": [{"CidrIp": "0.0.0.0/0"}]}
    # if it exists, just return the id
    security_group_name = get_name_hash(security_group_prefix, inbound_rule, max_length=255)
    try:
        security_group = ec2_client.describe_security_groups(GroupNames=[security_group_name])
        return security_group["SecurityGroups"][0]["GroupId"]
    except ec2_client.exceptions.ClientError:
        pass

    security_group = ec2_client.create_security_group(
        GroupName=security_group_name,
        Description="Security group swe rex",
        VpcId=vpc_id,
        TagSpecifications=[
            {
                "ResourceType": "security-group",
                "Tags": [{"Key": "origin", "Value": "swe-rex-deployment-auto"}],
            },
        ],
    )
    security_group_id = security_group["GroupId"]

    # Add an inbound rule to allow traffic on the port
    ec2_client.authorize_security_group_ingress(GroupId=security_group_id, IpPermissions=[inbound_rule])
    return security_group_id


def run_fargate_task(
    command: list[str],
    name: str,
    task_definition_arn: str,
    subnet_id: str,
    security_group_id: str,
    cluster_arn: str,
    vcpus: int = 1,
    memory: int = 2048,
    ephemeral_storage: int = 21,
    **fargate_args,
) -> str:
    run_task_args = {
        "taskDefinition": task_definition_arn,
        "launchType": "FARGATE",
        "cluster": cluster_arn,
        "networkConfiguration": {
            "awsvpcConfiguration": {
                "subnets": [subnet_id],
                "securityGroups": [security_group_id],
                "assignPublicIp": "ENABLED",
            }
        },
        "propagateTags": "TASK_DEFINITION",
    }
    overrides = {
        "containerOverrides": [
            {
                "name": name,
                "command": command,
            },
        ],
        "cpu": f"{vcpus}vCPU",
        "memory": str(memory),
        "ephemeralStorage": {
            "sizeInGiB": ephemeral_storage,
        },
    }
    if fargate_args:
        overrides.update(fargate_args)

    if overrides:
        run_task_args["overrides"] = overrides

    ecs_client = boto3.client("ecs")
    response = ecs_client.run_task(
        **run_task_args,
        tags=[{"key": "origin", "value": "swe-rex-deployment-auto"}],
    )
    return response["tasks"][0]["taskArn"]


def get_public_ip(task_arn: str, cluster_arn: str) -> str:
    ecs_client = boto3.client("ecs")
    task_details = ecs_client.describe_tasks(cluster=cluster_arn, tasks=[task_arn])
    eni_id = task_details["tasks"][0]["attachments"][0]["details"][1]["value"]
    ec2_client = boto3.client("ec2")
    eni_details = ec2_client.describe_network_interfaces(NetworkInterfaceIds=[eni_id])
    return eni_details["NetworkInterfaces"][0]["Association"]["PublicIp"]


def get_cloudwatch_log_url(task_arn: str, task_definition: dict, container_name: str, region: str = "us-east-2") -> str:
    """
    Get the CloudWatch log URL for a running task.

    Args:
        task_arn (str): The ARN of the running task.
        task_definition (dict): The task definition for the running task.
        container_name (str): The name of the container in the task definition.

    Returns:
        str: The CloudWatch log URL.
    """
    container_def = task_definition["containerDefinitions"][0]
    log_config = container_def["logConfiguration"]["options"]
    log_group = log_config["awslogs-group"]
    task_id = task_arn.split("/")[-1]
    log_stream = f"{log_config['awslogs-stream-prefix']}/{container_name}/{task_id}"

    # Use %25 instead of $25 for encoding
    encoded_log_group = quote(log_group, safe="")
    encoded_log_stream = quote(log_stream, safe="")

    return (
        f"https://{region}.console.aws.amazon.com/cloudwatch/home"
        f"?region={region}"
        f"#logsV2:log-groups/log-group/{encoded_log_group}"
        f"/log-events/{encoded_log_stream}"
    )



from pathlib import PurePath
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, model_validator

from swerex.deployment.abstract import AbstractDeployment


class LocalDeploymentConfig(BaseModel):
    """Configuration for running locally."""

    type: Literal["local"] = "local"
    """Discriminator for (de)serialization/CLI. Do not change."""

    model_config = ConfigDict(extra="forbid")

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.local import LocalDeployment

        return LocalDeployment.from_config(self)


class DockerDeploymentConfig(BaseModel):
    """Configuration for running locally in a Docker or Podman container."""

    image: str = "python:3.11"
    """The name of the container image to use."""
    port: int | None = None
    """The port that the container connects to. If None, a free port is found."""
    docker_args: list[str] = []
    """Additional arguments to pass to the container run command. If --platform is specified here, it will be moved to the platform field."""
    startup_timeout: float = 180.0
    """The time to wait for the runtime to start."""
    pull: Literal["never", "always", "missing"] = "missing"
    """When to pull container images."""
    remove_images: bool = False
    """Whether to remove the image after it has stopped."""
    python_standalone_dir: str | None = None
    """The directory to use for the python standalone."""
    platform: str | None = None
    """The platform to use for the container image."""
    remove_container: bool = True
    """Whether to remove the container after it has stopped."""
    container_runtime: Literal["docker", "podman"] = "docker"
    """The container runtime to use (docker or podman)."""

    type: Literal["docker"] = "docker"
    """Discriminator for (de)serialization/CLI. Do not change."""

    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="before")
    def validate_platform_args(cls, data: dict) -> dict:
        if not isinstance(data, dict):
            return data

        docker_args = data.get("docker_args", [])
        platform = data.get("platform")

        platform_arg_idx = next((i for i, arg in enumerate(docker_args) if arg.startswith("--platform")), -1)

        if platform_arg_idx != -1:
            if platform is not None:
                msg = "Cannot specify platform both via 'platform' field and '--platform' in docker_args"
                raise ValueError(msg)
            # Extract platform value from --platform argument
            if "=" in docker_args[platform_arg_idx]:
                # Handle case where platform is specified as --platform=value
                data["platform"] = docker_args[platform_arg_idx].split("=", 1)[1]
                data["docker_args"] = docker_args[:platform_arg_idx] + docker_args[platform_arg_idx + 1 :]
            elif platform_arg_idx + 1 < len(docker_args):
                data["platform"] = docker_args[platform_arg_idx + 1]
                # Remove the --platform and its value from docker_args
                data["docker_args"] = docker_args[:platform_arg_idx] + docker_args[platform_arg_idx + 2 :]
            else:
                msg = "--platform argument must be followed by a value"
                raise ValueError(msg)

        return data

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.docker import DockerDeployment

        return DockerDeployment.from_config(self)


class ModalDeploymentConfig(BaseModel):
    """Configuration for running on Modal."""

    image: str | PurePath = "python:3.11"
    """Image to use for the deployment."""

    startup_timeout: float = 180.0
    """The time to wait for the runtime to start."""

    runtime_timeout: float = 60.0
    """Runtime timeout (default timeout for all runtime requests)
    """

    deployment_timeout: float = 3600.0
    """Kill deployment after this many seconds no matter what.
    This is a useful killing switch to ensure that you don't spend too 
    much money on modal.
    """

    modal_sandbox_kwargs: dict[str, Any] = {}
    """Additional arguments to pass to `modal.Sandbox.create`"""

    type: Literal["modal"] = "modal"
    """Discriminator for (de)serialization/CLI. Do not change."""

    install_pipx: bool = True
    """Whether to install pipx with apt in the container.
    This is enabled by default so we can fall back to installing swe-rex
    with pipx if the image does not have it. However, depending on your image,
    installing pipx might fail (or be slow).
    """

    model_config = ConfigDict(extra="forbid")

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.modal import ModalDeployment

        return ModalDeployment.from_config(self)


class FargateDeploymentConfig(BaseModel):
    """Configuration for running on AWS Fargate."""

    image: str = "python:3.11"
    port: int = 8880
    cluster_name: str = "swe-rex-cluster"
    execution_role_prefix: str = "swe-rex-execution-role"
    task_definition_prefix: str = "swe-rex-task"
    log_group: str | None = "/ecs/swe-rex-deployment"
    security_group_prefix: str = "swe-rex-deployment-sg"
    fargate_args: dict[str, str] = {}
    container_timeout: float = 60 * 15
    runtime_timeout: float = 60

    type: Literal["fargate"] = "fargate"
    """Discriminator for (de)serialization/CLI. Do not change."""

    model_config = ConfigDict(extra="forbid")

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.fargate import FargateDeployment

        return FargateDeployment.from_config(self)


class RemoteDeploymentConfig(BaseModel):
    """Configuration for `RemoteDeployment`, a wrapper around `RemoteRuntime` that can be used to connect to any
    swerex server.
    """

    auth_token: str
    """The token to use for authentication."""
    host: str = "http://127.0.0.1"
    """The host to connect to."""
    port: int | None = None
    """The port to connect to."""
    timeout: float = 0.15

    type: Literal["remote"] = "remote"
    """Discriminator for (de)serialization/CLI. Do not change."""

    model_config = ConfigDict(extra="forbid")

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.remote import RemoteDeployment

        return RemoteDeployment.from_config(self)


class DummyDeploymentConfig(BaseModel):
    """Configuration for `DummyDeployment`, a deployment that is used for testing."""

    type: Literal["dummy"] = "dummy"
    """Discriminator for (de)serialization/CLI. Do not change."""

    model_config = ConfigDict(extra="forbid")

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.dummy import DummyDeployment

        return DummyDeployment.from_config(self)


class DaytonaDeploymentConfig(BaseModel):
    """Configuration for Daytona deployment."""

    api_key: str = Field(default="", description="Daytona API key for authentication")
    target: str = Field(default="us", description="Daytona target region (us, eu, etc.)")
    port: int = Field(default=8000, description="Port to expose for the SWE Rex server")
    container_timeout: float = Field(default=60 * 15, description="Timeout for the container")
    runtime_timeout: float = Field(default=60, description="Timeout for the runtime")
    image: str = Field(default="python:3.11", description="Image to use for the sandbox")

    def get_deployment(self) -> AbstractDeployment:
        from swerex.deployment.daytona import DaytonaDeployment

        return DaytonaDeployment.from_config(self)


DeploymentConfig = (
    LocalDeploymentConfig
    | DockerDeploymentConfig
    | ModalDeploymentConfig
    | FargateDeploymentConfig
    | RemoteDeploymentConfig
    | DummyDeploymentConfig
    | DaytonaDeploymentConfig
)
"""Union of all deployment configurations. Useful for type hints."""


def get_deployment(
    config: DeploymentConfig,
) -> AbstractDeployment:
    return config.get_deployment()
